-----------------------------------------------------------------------
-- To do and/or ideas
--
-- From Rabbit:
-- * Should probably print the current damage when auto-resetting.
-- * Add % display, that would show the percentage of the total damage in the
-- data table each player had contributed to, maybe replace the current damage
-- number or add it as a 4th column.
-- * Parse damage messages from swstats and damagemeters.
-- * Get someone to make us an icon.
-- * Automatically output our damage done and position to chat when a boss has
-- been killed; this is easy with Babble-Boss, CHAT_MSG_HOSTILE_DEATH and
-- UnitClassification.
-- * Maybe add an auto-reset for boss kill as well.
--
-- From Nymbia:
-- * Pick up the slack of users without the addon by splitting up the members
-- without the addon among those with it.  ie, in a 25 man raid where 10 users
-- have Violation, each of them is assigned to parse the damage of either 1 or
-- 2 other raid members.
-- * Possibly keep two running totals:  'this fight' and 'overall', to allow
-- for statistics to be displayed for a single encounter without requiring
-- constant resets.
--
-- From <person>:
-- * Brew my coffee when I win the damage meter contest after a boss fight.
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation")

L:RegisterTranslations("enUS", function() return {
	["Resetting data as requested by %s."] = true,
	["Reset"] = true,
	["Options for when to reset the data."] = true,
	["Reset now"] = true,
	["Immediately resets your data."] = true,
	["Reset group"] = true,
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = true,
	["Leader resets"] = true,
	["Accept reset requests from raid leaders or officers."] = true,
	["Auto-reset"] = true,
	["Entering combat"] = true,
	["Automatically reset the data when entering combat."] = true,
	["Leaving combat"] = true,
	["Automatically reset the data when leaving combat."] = true,

	["Show data"] = true,
	["Shows your own data on the FuBar panel, if available."] = true,

	["No data gathered yet."] = true,
	["|cffeda55fClick|r to reset."] = true,
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")
local crayon = AceLibrary("Crayon-2.0")

local playerName = nil

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		self[key] = "|cff" .. BC:GetHexColor(obj.class) .. key .. "|r"
		return self[key]
	end
})

local data = nil
local ownDataownDataDirty = nil
local playerPosition = nil

local db = nil

local _G = getfenv(0)

-----------------------------------------------------------------------
-- Local heap
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select('#', ...) do
				table.insert(t, select(i, ...))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	args = {
		text = {
			type = "toggle",
			name = L["Show data"],
			desc = L["Shows your own data on the FuBar panel, if available."],
			get = function() return Violation.db.profile.showDamageText end,
			set = function(v)
				Violation.db.profile.showDamageText = v
				Violation:UpdateText()
			end,
			disabled = function() return not FuBar end,
			order = 100,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = function() Violation:Reset() end,
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = function() Violation:ResetGroup() end,
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = function() return Violation.db.profile.leaderReset end,
					set = function(v) Violation.db.profile.leaderReset = v end,
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return Violation.db.profile.gainCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						Violation.db.profile.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return Violation.db.profile.leaveCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						Violation.db.profile.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 101,
		},
	},
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceComm-2.0", "FuBarPlugin-2.0", "Parser-3.0")
local Violation = Violation

Violation.hasIcon = "Interface\\Icons\\Spell_ChargePositive"

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 28447 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB")
	self:RegisterDefaults("profile", {
		gainCombat = nil,
		leaveCombat = nil,
		leaderReset = true,
		showDamageText = nil,
	})

	self.OnMenuRequest = options
	self.hasNoColor = true

	self:RegisterChatCommand({ "/violation" }, self.OnMenuRequest, "VIOLATION")

	playerName = UnitName("player")
end

function Violation:OnEnable()
	self:Reset()

	self:RegisterParserEvent({
		eventType = "Damage",
		sourceID = "player",
	}, "OnCombatEvent")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")
	self:RegisterComm("Violation", "GROUP", "OnComm")

	self:ScheduleRepeatingEvent("DamageBroadcast", self.Broadcast, 5, self)

	db = self.db.profile

	self:RegisterEvent("CHAT_MSG_ADDON", "Listen")

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end
end

function Violation:Reset()
	if type(data) == "table" then
		for k, v in pairs(data) do
			data[k] = nil
		end
	else
		data = {}
	end
	data[playerName] = 0
	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

function Violation:OnClick()
	self:Reset()
end

local function Sorter(a, b)
	return a[2] > b[2]
end

function Violation:OnTooltipUpdate()
	local tbl = new()
	for k, v in pairs(data) do
		if type(v) == "number" and v > 0 then
			table.insert(tbl, new(k, v))
		end
	end

	if #tbl > 0 then
		local cat = tablet:AddCategory(
			"columns", 3,
			"child_justify", "LEFT",
			"child_justify2", "LEFT",
			"child_justify3", "RIGHT",
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text3R", 1,
			"child_text3G", 1,
			"child_text3B", 1
		)

		table.sort(tbl, Sorter)

		local playerShown = nil
		for i, v in ipairs(tbl) do
			if i > 10 then break end -- Only show the top 10.
			cat:AddLine("text", tostring(i) .. ".", "text2", coloredNames[v[1]], "text3", tostring(v[2]))
			if v[1] == playerName then
				playerShown = true
				playerPosition = i
			end
		end

		if not playerShown and data[playerName] > 0 then
			cat:AddLine("text", "?", "text2", coloredNames[playerName], "text3", tostring(data[playerName]))
			playerPosition = "?"
		end
	else
		local cat = tablet:AddCategory("columns", 1)
		cat:AddLine("text", L["No data gathered yet."])
	end

	-- Cleanup.
	tbl = del(tbl)

	tablet:SetHint(L["|cffeda55fClick|r to reset."])
end

function Violation:OnTextUpdate()
	if db and db.showDamageText and type(data[playerName]) == "number" and data[playerName] > 0 then
		self:SetText((type(playerPosition) ~= "nil" and tostring(playerPosition) .. ". " or "") .. data[playerName])
	else
		self:SetText(self.name)
	end
end

-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------
function Violation:Broadcast()
	if ownDataDirty == 0 or not ownDataDirty then return end
	self:SendCommMessage("GROUP",ownDataDirty)
	ownDataDirty = 0
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		error("Only the group leader or an officer can do that.")
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
end

function Violation:OnComm(prefix, sender, distribution, input)
	if distribution ~= "GROUP" then return end
	if tonumber(input) then
		data[sender] = (data[sender] or 0) + tonumber(input)
		self:ScheduleEvent("DisplayUpdate", self.UpdateDisplay, 5, self)
	elseif type(input) == "string" then
		if input == "LeaderReset" then
			local obj = roster:GetUnitObjectFromName(sender)
			if obj.rank > 0 then
				self:Print(string.format(L["Resetting data as requested by %s."], sender))
				self:Reset()
			end
		end
	end
end

function Violation:Listen(prefix, message, distribution, sender)
	if distribution ~= "RAID" then return end
	if prefix ~= "DamageMeters" then return end

	local formatStr = "SYNC_12_ (%w+) (%d+)"
	local _, _, player, dmg = string.find(message, formatStr)

	if player and player ~= playerName and player == sender then
		data[player] = tonumber(dmg)
		self:ScheduleEvent("DisplayUpdate", self.UpdateDisplay, 5, self)
	end
end

-----------------------------------------------------------------------
-- Combat log parsing
-----------------------------------------------------------------------
function Violation:OnCombatEvent(t)
	assert(t.sourceName == "player", "Not supposed to get combat events from people except ourselves.")
	if t.amount then
		data[playerName] = data[playerName] + t.amount
		ownDataDirty = (ownDataDirty or 0) + t.amount
		self:UpdateDisplay()
	end
end

